\     Turtle-Logo in Forth, Copyright (c) 2016-2018, Assad Ebrahim, Mathematical Science & Technologies
\          FREE download from: http://www.mathscitech.org/articles/turtle-logo-forth
\          Email assad.ebrahim@alum.swarthmore.edu for comments, bug-reports, feature-requests

\     GNU License - Users are permitted free re-use provided:                          
\           1/ Source code for all modified versions are made freely available under the GNU License            
\           2/ Attribution and this header is preserved (lines 1-7)

\     Description: This is a Turtle logo program for Windows operated entirely with a few keys.  
\                  It has several advanced features, including macro recording and replay.
\                  It is written in 950 lines of Forth code for the F-PC Forth for DOS.
\                  It runs within a freely available DOS emulator (vDOSplus) that is included in this package.
\                  This package with Windows installer is available at the website:  
\                        http://www.mathscitech.org/articles/turtle-logo-forth
\                  TURTLE LOGO IS INSTALLED TO: C:\TOTALCMD\TURTLE-LOGO\

\     Turnkey Startup: Run TurtleLogo.exe (vDosPlus) which in turn calls "turtle_logo.bat".
\                      Further instructions are provided in the _readme.txt

\     Developers Startup: To modify the program, edit turtle.fs and run it in F-PC (Forth FPC) using command:
\                                   fload turtle.fs, then: turtle-go
\                        To package a modified release, recompile F-PC with FMETA to create F-PC with turtle.fs embedded in it (turnkey application)
\                        F-PC is included in this package, or can be downloaded from: ftp://ftp.taygeta.com/pub/Forth/Compilers/native/dos/FPC/fpc36.zip
\                        Instructions included in docs: Fpc\doc\ting_1989_FPC_UsersManual\Chapter9_Rebuilding_FPC.txt

\ === REVISION HISTORY
: rev# ." 1.6.4042" ;

\ === INSTRUCTIONS
: welcome ( -- )
    ." WELCOME TO TURTLE LOGO! (v" REV# ." ), (C) 2016-2018, ASSAD EBRAHIM"  \ All Rights Reserved
    cr ." QUICK START:"
    cr ."   (1) To MOVE Turtle, use the arrow keys [FWD, LEFT, RIGHT] and [BACKSPACE]."
    cr ."     [FWD] moves one step along current heading and paints pixel behind Turtle."
    cr ."     [LEFT]/[RIGHT] turns Turtle's heading 45 deg counter-/clock- wise."
    cr ."     [BACKSPACE] backs up & erases."                   
    cr ."     [H] sends Turtle to home position at bottom left of screen."
    cr ."     Note that movement wraps around the edges of the screen to the other side!
    cr ."   (2) To CHANGE Turtle's paint modes... "
    cr ."     ['] toggles pen-up to move without painting and pen-down to draw again."
    cr ."     [C/X] cycles pen through 16 color choices (fwd/back)."
    cr ."     [;] toggles Turtle's visibility (hide/show)
    cr ."     [ puts Turtle into text mode (now type)   ] returns Turtle to draw mode."
    cr ."     [ESC] clears screen (once you confirm). [Z] activates cloaking stealth mode!"
    cr ."   (3) You can TEACH Turtle skills! He will remember & repeat."  \ RECORD/REPLAY macros..."
    cr ."     [N] cycles through skills 1-9."                                             \ Note, it will be EXACTLY what you told him.
    cr ."     [T] starts/ends TEACHING Turtle that SKILL. (Turtle can learn 9 skills.)"   \ skill=diary.  can record nested macros, but not recursive macros
    cr ."     [1-9] gets Turtle to replay that skill. [0] shows all skills he knows."
    cr ."   (4) OTHER things you can do..."
    cr ."     [S] saves the screen; [U] undoes all actions to the last save point"
    cr ."     [G] greets Turtle (it will greet you back!)"
    cr ."     [?] shows this HELP screen (your drawing is safe!)"
    cr ."     [Q] quits the program (after confirming)"
    cr ." HAPPY TURTLE-ING!"
\    cr ." For more downloads: http://www.mathscitech.org/articles/turtle-logo-forth"
    cr ;

\ === Interactivity / Scaffolding
  $1B        constant   esc-key          \ 27d
: pause-key  ( -- )     ." PRESS ANY KEY TO CONTINUE... " key drop ;
: pause-esc  ( -- t/f ) ." PRESS any key for next skill; ESC to return to drawing..." key esc-key =  ;

\ === Numeric conversion and Tests
: invert      ( n -- n' )       not   ;                           \ bitwise logical not (invert odd)
: is_between? ( n a b -- n t/f ) rot dup rot > -rot > or invert ;  \ not ( n>b or n<a )
: is_char?    ( asc -- asc t/f ) $20  $7E is_between? ;
: a2d         ( asc -- digit )  '0' - ;                           \ ascii to digit
: d2a         ( digit -- asc )  '0' + ;                           \ digit to ascii

\ === Bit manipulation
: shl4      ( val -- val )        $10 * ;                 \ raise nybble
: shr4      ( val -- val )        $10 / ;                 \ lower nybble

\ === Memory
: bytes     ( n -- n )  ;                                 \ 1x, for documentation
: cells     ( num -- num*wsz )        2 *  ;                  \ 16-bit word length on 8086 chip
: c@+       ( adr -- adr' val )       dup 1 + swap c@ ;       \ fetch/store char/word and advance pointer
: @+        ( adr -- adr' val )       dup 1 cells + swap @ ;  
: c!+       ( val adr -- adr' )       dup 1 + -rot c! ;       
: !+        ( val adr -- adr' )       dup 1 cells + -rot ! ;  
: cmove+    ( c-adr dest n -- dest' )  >r tuck r@ cmove r> + ;     \ copy byte sequence/string (c-addr, n bytes) to dest adr, returning new position
: attach+        ( dest-adr c-adr nchar -- dest-adr' )   >r swap r> cmove+ ;  \ attach to next memory point, returning new position
: get-fw-string ( base-adr fw-len idx -- c-adr fw-len ) over * rot + swap ;  \ returns idx-th fixed width string

\ === Strings
$22   constant    dquote                  \ double quote
: s"  ( <inputstream"> -- c-adr nchar )  dquote word c@+ pad swap 2dup >r >r cmove r> r> ;   \ make next quote-terminated string into quoted string
: q2c-string ( c-adr nlen -- ) dup c, here swap dup bytes allot cmove ;  \ quoted string to counted string
    
\ === Files
: file-write   ( Addr len filename --- )   \  write len bytes from addr into the specified file
    count seqhandle+ ">handle
    seqhandle+ HDELETE   DROP
    seqhandle+ HCREATE   ABORT" Save Create ERR!"
    seqhandle+ HWRITE 0= ABORT" Save Write  ERR!"
    seqhandle+ HCLOSE    ABORT" Save Close  ERR!"  ;

\ Filenames - program writes to file screen and diary when user saves [s] and when program quits [q]
create fscr-save s" .saved-screen"  q2c-string
create fskl-save s" .saved-skills"  q2c-string
create fscr-quit s" .quit-screen"   q2c-string
create fskl-quit s" .quit-skills"   q2c-string

\ === Heading through 8 Compass Directions
\ Heading is an integer 0 - 7 which indexes into an array of heading symbols that show which direction turtle is heading
\ Turtle cycles mod-8 through the directions as cw (turning right)  N > NE > E > SE > S > SW > W > NW  or ccw (turning left)
create heading 1 bytes allot  \  VARIABLE - holds turtle heading, integer 0 to 7    
: set-heading  ( h# -- )    heading c! ;
: get-heading  ( -- h# )    heading c@ ;
: init-heading ( -- )    0  set-heading ;  \ default is N=0

: create-heading-symbol
create ( <name> -- ) '^' c, '/' c, '>' c, '\' c, 'v' c, '/' c, '<' c, '\' c,   \ in cw cycle order (N,NE,E,...)
    does> ( h# -- asc )  + c@ ;
create-heading-symbol  HEADING-SYMBOL

: create-heading-string
create ( <name> -- ) here 8 cells allot
  " N " attach+  " NE" attach+   " E " attach+   " SE" attach+
  " S " attach+  " SW" attach+   " W " attach+   " NW" attach+
        drop
does> ( h# base-adr -- c-adr nchar ) 2 rot get-fw-string ;
create-heading-string  HEADING-STRING


\ =========  16 Colors
16               constant   num-colors
num-colors       constant   penup-col-index
num-colors 1 +   constant   num-color-str    \ 1 more for penup string
 0      constant  black       1      constant  navy      2      constant  green
 3      constant  teal        4      constant  red       5      constant  purple 
 6      constant  brown       7      constant  silver    8      constant  slate 
 9      constant  blue       10      constant  lime     11      constant  aqua  
 12     constant  orange     13      constant  pink     14      constant  yellow
 15     constant  white 
 \ alternate spellings
 silver constant  grey        silver constant  gray       
 slate  constant  darkgrey    slate  constant  darkgray

\ ===  Color Names
\ Color comments: navy=dark blue, teal=aqua, red=maroon, silver=grey, slate=darkgrey, blue=lt blue, lime=lt green, aqua=lt aqua/cyan, pink=magenta/fuschia
6        constant    strlen-col
: create-color-names
create ( -- ) here num-color-str strlen-col * bytes allot
" Black " attach+      " Navy  " attach+      " Green " attach+
" Teal  " attach+      " Red   " attach+      " Purple" attach+
" Brown " attach+      " Silver" attach+      " Slate " attach+
" Blue  " attach+      " Lime  " attach+      " Aqua  " attach+
" Orange" attach+      " Pink  " attach+      " Yellow" attach+
" White " attach+      " PEN-UP" attach+
  drop
does> ( idx base-adr -- c-adr nchar ) strlen-col rot get-fw-string ;   \ return idx-th fixed width color-name
create-color-names COLOR-NAMES 


\ == COLOR STATE VARIABLES 
create border-color 1 bytes allot 
create screen-color 1 bytes allot  \ VARIABLE - holds canvas color codes 0..15
create turtle-color    1 bytes allot  \ VARIABLE -- pencolor is foreground color, 0..15
create pen-color 1 bytes allot  \ VARIABLE - holds pen (background) color --- NOTE: this is the color the turtle draws
create prev-pen-color 1 bytes allot  \ VARIABLE

: set-border-color ( color -- ) border-color c! ;
: get-border-color ( -- color ) border-color c@ ;

: set-screen-color ( color -- )   screen-color c! ;
: get-screen-color ( -- color )   screen-color c@ ;

: set-turtle-color ( bgcolor -- ) turtle-color c! ;
: get-turtle-color ( -- bgcolor ) turtle-color c@ ;

\ === Turtle/Pen color mapping (foreground/background).
\ The design chooses an aesthetic pen color (foreground) for each pen color (background) 
\ Rationale: Color of pen (fgcolor) should contrast sufficiently with background to discern but not so much as to confuse on what color the turtle will draw.
\ Initial Method (no longer used): Theory of color contrasts using contrast ratio (contrast checker tool)
\               http://www.worqx.com/color/color_contrast.htm
\               http://webaim.org/resources/contrastchecker/
\               http://juicystudio.com/services/luminositycontrastratio.php#specify

 
 \ === Color encoding (16-color text mode): colorcode is one byte with background color in hi nybble and foreground color in low nybble
: fgbg2col ( fgcolor bgcolor -- colorcode )  shl4 + ;
: col2fgbg ( colorcode -- fgcolor bgcolor )  dup $0F and  ( fgcolor )  swap shr4  ( bgcolor ) ;

\ fgbg lookup table is indexed by  bgcolor (pen/cursor) and provides as lookup the corresponding fgcolor -- order matters
: create-fgbg-lookup
create
   \ fg           \ bg (index)
   silver   c,    \ black    
   aqua     c,    \ navy     
   lime     c,    \ green
   aqua     c,    \ teal     
   pink     c,    \ red      
   white    c,    \ purple   
   lime     c,    \ brown    
   slate    c,    \ silver   
   silver   c,    \ slate    
   aqua     c,    \ blue     
   green    c,    \ lime     
   navy     c,    \ aqua     
   yellow   c,    \ orange   
   red      c,    \ pink     
   orange   c,    \ yellow   
   silver   c,    \ white     
\ Note: penup is handled with specific penup color instructions
does> ( fg -- ) + c@ set-turtle-color ;
create-fgbg-lookup  SET-OPTIM-FG-COLOR

: set-pen-color    ( c -- )  dup  pen-color c! set-optim-fg-color ;         \ selecting background automatically changes turtle-color so it can be seen
: get-pen-color    ( -- c )  pen-color c@ ;
: switchto-penup-color  ( -- ) lime set-pen-color white turtle-color c! ;   \ gives penup a distinct appearance

: save-prev-pen-color    ( -- )   get-pen-color prev-pen-color c! ;
: restore-prev-pen-color ( -- c ) prev-pen-color c@ set-pen-color ;      \ ??? TESTER REVIEW search previous comment

: color-change-ex ( 1/-1 -- )  get-pen-color  +  16  mod  set-pen-color ;  \ +1/-1 cycle fwd/bk to next color of 16


\ =========== Graphics.
\ VGA in DOS is a quick way to produce graphics because in DOS it is easy to access screen pixels (peek/poke) directly via VGA card memory.  
\ This is no \ longer the case in modern windows which requires going through a graphics library such as SDL or OpenGL.
\ This program is written F-PC (16-bit Forth) because F-PC makes it easy to write 8086 assembly code, making VGA on DOS straightforward. 
\ Modern Windows computer?  No problem, use a DOS emulator (vDosPlus recommended or DOSBox).

\ Standard text mode graphics works with 80 columns and 25 rows (2000 cells, indices x:0..79, y:0..24 ).  
\ Implementation of each row is with 160 data columns, so 4000 cells.  
\ Drawing works correctly only on even columns so in practice the 80 column indices are  multiples of 2, i.e. 0,2,4,...,158.
\ Pixel locations are specified in linear order from 0 to 3998d ($F9E in hex) with the final 3999d cell not visible (even pixel only issue).

\ Two representations are possible: modified Cartesian x,y as across and *down* vs. Matrix: i,j for row and column (down/across).
\ THIS PROGRAM uses Cartesian (x,y / across/down).

\ We have THREE coordinate systems: 
\     USER COORDINATES - what the user can access on the screen, i.e. excluding menus, borders, etc., with 0,0 as the origin point for the user's cursor.
\           i.e. within a window, (i.e. a subset of the 80x25 full-screen)
\     SCREEN COORDINATES - where this all happens on the screen, with (0,0) at top left corner of the screen
\     PHYSICAL COORDINATES - the underlying implementation / linear data array against which the drawing pixels must be referenced.
\   xy2lin converts from USER coordinates from physical position in linear array.  lin2xy goes from physical to user.
\   xy2lin-screen converts from SCREEN coordinates to physical.  No need for back-conversion.

\ Example - FPC uses rows 0-1 (1st line is Forth bar, and 2nd line is erased after command scrolls screen up), so to see a pixel within F-PC start at linear position 320 (start of row 3)
\ FPC uses columns 69-79 of the first 12 rows for dictionary, with the full 80 cols starting from row 13.

\   References:
\   Video programming:  http://ece.wpi.edu/~wrm/Courses/EE3803/Labs/roehrl.html
\   Video mode:         http://www.wagemakers.be/english/doc/vga

\ PC Video Mode (BIOS settings)
$1 constant text40x25-8color
$3 constant text80x25-8color            \ text mode (8 color)
$8 constant text80x25-16color           \ default F-PC mode, colsxrows, each character is 8x8 pixels (640 pixels x 200)
$4 constant graphCGA-320x200-8color     \ standard resolution
$10 constant graphEGA-640x350-16color   \ high resolution
$12 constant graphVGA-640x480-16color   \ default VGA graphics mode

code _set-video-mode ( mode -- )   \ calls PC BIOS instruction to set desired video mode
    pop ax          \ AL=video mode 
    mov ah, # 0     \ AH=set video mode (BIOS instruction)
    int $10		    \ DOS interrupt: call BIOS instruction
next end-code

: text-video-mode     ( -- ) text80x25-16color        _set-video-mode ;
: graphics-video-mode ( -- ) graphVGA-640x480-16color _set-video-mode ;

\ === Physical (linear array) coordinates for text mode: 80x25 = 2000 visible cells.  160x25 = 4000 implementation cells.
  80  constant screen-cols
 160  constant physical-cols  \ note: odd slots are invalid display locations, so counter runs 0,2,4,..,178   (see graphics discussion above)
  25  constant screen-rows
   0  constant physical-zero
3998  constant physical-end    \ $F9E hex
\ From PHYSICAL to SCREEN coordinates
 0  constant top-menu-height
 3  constant bottom-menu-height
 1  constant buffer-thickness  \ frame all round

\ === USER area in SCREEN coordinates (0,0 is top left pixel of the screen)
0 top-menu-height + buffer-thickness +                  constant user-top
screen-rows  bottom-menu-height -  buffer-thickness - 1 -  constant user-bot
buffer-thickness                                        constant user-left
screen-cols buffer-thickness - 1 -                        constant user-right
user-bot user-top - 1 +                                 constant user-height
user-right user-left - 1 +                              constant user-width
2 1 - physical-cols *                                   constant row#3      \ USER COORDINATE STARTS HERE 3rd line ($140=320d) is one accessible in  user mode (1st is menu, 2nd is border)

\ === Linear offset in physical coordinates to 0,0 in user coordinates (top left pixel that the user can access, i.e. excluding menus and buffers)
user-top physical-cols * user-left 2 * + constant lin-offset

\ === Coordinate conversion
\ Screen coordinates are Cartesian ( x horizontal, y vertical), with graphics convention (0,0 is top left corner, y is down).  Not using matrix convention ( row i, col j).  
\ Physical Implementation is in linear array, row by row (aka physical coordinates)
\ User coordinates are subset of Screen coordinates, with offsets that prevent user from moving into forbidden areas (menus, etc.)

: xy2lin ( x y -- lin )  \ USER to physical coordinates
\ test: 0 0 xy2lin  = 320, is 3rd row down
    physical-cols *     \ y rows to linear array
    swap 2 *        \ x cols (0..79) to even x-slots (0,2,4,...158) since only evens are visible
    +               \ combine to get linear position
    lin-offset + ; \ plus linear offset to reference to user coordinates (0,0) 
: lin2xy ( lin -- x y )  \ physical to USER coordinates
    lin-offset -   \ remove offset
    physical-cols /mod ( x y ) 
    swap 2 /          \ to screen char coords (80 per row)
    swap ;
: xy2lin-screen ( x y -- lin )  physical-cols * swap 2 * + ; \ SCREEN to physical coordinates

: init-turtle-position ( -- x y ) 0 user-height 1 - ;
: print-location ( x y -- x y )   2dup ." (x: " swap . ." y: " . ." )" ;    \ TO DO: Add this to status bar

\ === BASE GRAPHICS - Painting to Screen- Low-level Character Routines
\ color information is encoded into a single byte for use in assembly language routine.  
\ color-code (e.g. $F4) holds background color (bgcolor) in high nybble ($F is white) and foreground (text) color in low nybble ($4 is red)
\ pixel position is in linear encoded form, i.e. row/col are encoded within a single wrap-around array of screen positions

code poke-pixel  ( linear-location  ascii-code  color-code  -- )  \ writes pixel to screen via direct memory address on vga card
                                                                  \ pixel is at physical position in linear array with given colors and ascii character.
\ test: 10 10 xy2lin '*' red yellow fgbg2col poke-pixel
    pop ax                  \ load AL with color-code (foreground & background)
    pop cx                  \ load CL with ascii char
    mov ch, al              \ complete CX as text pixel: (color|char) in (CH|CL)
    pop bx                  \ load BX with screen location
    push ds                 \ preserve data stack
    mov ax, # $B800         \ load AX with graphics address B800 hex (vga memory base address for direct display to screen)
    mov ds, ax              \ set data segment to graphics address (cannot load literal directly, only via register)
    mov 0 [bx], cx          \ display pixel in desired location in graphics memory [B800:bx] (0 data segment offset)
    pop ds                  \ restore data stack
next end-code
\ test: 500 BL $FF poke-pixel
\ test: 500 42 228 
\ test: 320 42 228  
\ test: 77 20 xy2lin 42 228 

code peek-pixel   ( linear-location -- ascii-code color-code )  \ get pixel information from screen; if position is x y use xy2lin first
\ test: turtle-init 0 20 xy2lin peek-pixel
    pop bx              \ load BL with screen location
    push ds             \ preserve data stack
    mov ax, # $B800     \ load AX with graphics address (B800 hex)
    mov ds, ax          \ set data segment to graphics address
    mov cx, 0 [bx]      \ load CX with pixel data from desired location in graphics memory (B800:bx)  (CH|CL)=(color|char)
    pop ds              \ restore data stack
    mov ax, # 0         \ clear AX
    mov al, ch          \ load AX with color-code
    mov ch, # 0         \ clear CH, leaving CL with ascii code
    push cx             \ return ascii code    
    push ax             \ return color-code
next end-code


\ === DRAWING STATE VARIABLES   note: boolean variables must be cells for logical true (-1) to work
create cloak-state 1 cells allot  \ BOOLEAN VARIABLE - true allows cloaking (movement without detection, i.e. invisibility mode)
: enable-cloaking ( -- ) true cloak-state c! ; 
: disable-cloaking ( -- ) false cloak-state c! ; 
: is-cloaking-on? ( -- t/f ) cloak-state c@ ; 

\ Show/Hide Turtle
create visible-state 1 cells allot  \ BOOLEAN VARIABLE - turtle visibility state -- true if visible, false if hidden
: set-visible ( -- ) true visible-state c! ;
: set-hidden  ( -- ) false visible-state c! ;
: is-visible? ( -- t/f) visible-state c@ ;

\ Turtle's Pen-mode 
create penup-state      1 cells allot  \ BOOLEAN VARIABLE - true if pen is up else false
create prev-penup-state 1 cells allot
: set-penup   ( -- ) true penup-state c!  ; 
: clear-penup ( -- ) false penup-state c! ; 
: is-penup?   ( -- t/f ) penup-state c@   ; 
: save-prev-penup-state  ( -- ) is-penup? prev-penup-state c! ;  \ used for cloaking toggle to not clobber state
: restore-prev-penup-state ( -- ) prev-penup-state c@ penup-state c! ;

\ Turtle's Text mode
create text-mode 1 cells allot     \ BOOLEAN VARIABLE - true if in text mode else false in draw mode
: set-text-mode ( -- ) true text-mode c! ;
: set-draw-mode ( -- ) false text-mode c! ; 
: is-text-mode? ( -- t/f ) text-mode c@ ; 

\ displaying strings to any location on screen
: write-char+ ( colorcode linloc char-adr  -- colorcode linloc' char-adr' ) 
        over 2 + -rot ( col l' l c-addr )       \ save next location (2 bytes ahead - see graphics explanation: 80 screen positions, 160 physical positions)
        c@+ rot swap ( col l' c' l ch )    \ get char and save next char position
        4 pick  ( col l' c' l char col )    \ setup stack
        poke-pixel ;
\ test: 0 20 xy2lin    red lime fgbg2col   s" Hello!"   drop   write-char+
: write-string ( linloc char-addr nchar fg bg -- ) rot >r fgbg2col -rot r> 0 do write-char+ loop 3drop ;
\ test: 0 20 xy2lin    s" Hello!"   white green   write-string


\ === DERIVED GRAPHICS == 
: fill-screen ( color -- )  \ color (0-15) or color-name ( black, white, yellow, red, ... )
    physical-zero physical-end rot dup fgbg2col -rot                     \ create color code 
    BL -rot                                   \ fill with blanks    
    over - 2/ 1 +                             \ length of physical array to fill but only the evens (odd cells go green in this video mode) 
( colorfgbg BL start-loc length )             
    0 do ( colorfgbg BL start-loc )           \ Only start-loc needs to change, rest are constant.  Looped instructions: 7.
        >r 2dup r@ -rot swap poke-pixel r> 2 +     \ increment by 2 as only even cells are visible
    loop 3drop ;  
: black-screen  ( -- ) black fill-screen ;

0  constant top-border-row
21 constant bot-border-row

: line-pixel ( color * x y -- color * ) xy2lin-screen ( linloc ) rot >r r@ dup fgbg2col ( color ) bl ( asc ) swap poke-pixel r> swap ;
: hline-draw ( color x1 x2 y -- ) -rot swap do ( color y ) i over line-pixel loop 2drop ;   \ draw horizontal line at y from x1 to x2 of color
: vline-draw ( color x y1 y2 -- ) swap do ( color x ) dup i line-pixel loop 2drop ;   \ draw vertical line at x from y1 to y2 of color
: borders-draw ( -- )  \ all coords are in physical space
    get-border-color  0 80  top-border-row                 hline-draw
    get-border-color  0 80  bot-border-row                 hline-draw
    get-border-color  0     top-border-row bot-border-row  vline-draw
    get-border-color 79     top-border-row bot-border-row  vline-draw ;


\ === PIXELS & PICTURE
\ A pixel is a 2-byte pair of ascii code, color-code, stored in that order.
\ Picture is a single array (nrow x ncol x 3 bytes) of pixels used to save/restore picture to/from memory.
\ Pixel variable is a temporary storage to hold 1 pixel information while user is deciding, in case the picture gets impacted (e.g. home, penup/down, etc.)
2   constant   pixel-sz
create         pixel  pixel-sz   bytes allot    \ variable - temp storage for 1 pixel
screen-cols    screen-rows *     constant   num-pix-scr
num-pix-scr    pixel-sz *        constant   pic-sz               

create  picture      pic-sz bytes allot                    \ array - memory block to hold full screen - holds working screen for use in preserving graphics
create  my-picture   pic-sz bytes allot                    \ holds screenshot when user selects save
: picture-clear ( pic-ptr --- )  pic-sz 0 fill ;        \ zero out memory
picture    picture-clear 
my-picture picture-clear

create xy-coord 2 bytes allot   \ variable - storage for x y coord (used by undo)
: save-xy ( x y -- x y ) 2dup swap xy-coord c!+ c! ;
: restore-xy ( x y -- x' y' ) 2drop xy-coord c@+ swap c@ ;  \ restore last saved x y

: pix!+ ( ascii-code color-code pix-ptr -- pix-ptr' ) >r swap r> c!+ c!+ ;       \ store pixel, preserving pixel order
: pix@+ ( pix-ptr -- ascii-code color-code pixl-ptr' ) c@+ swap c@+ swap ;       \ fetch pixel
: cache-pixel ( x y -- x y ) 2dup xy2lin peek-pixel pixel pix!+ drop ;
: uncache-pixel ( -- 1 2 )   pixel pix@+ drop ; 

\ === Turtle behaviour changing a pixel
: show-heading  ( x y h# -- )   -rot xy2lin swap HEADING-SYMBOL get-turtle-color get-pen-color fgbg2col poke-pixel ;  \ draw symbol on screen where turtle is
: paint-xy      ( x y -- x y )  2dup xy2lin BL get-pen-color dup fgbg2col poke-pixel ; \ test 3 12  \ after turtle has moved on (if pendown)
: restore-pixel ( x y -- x y )  2dup xy2lin uncache-pixel poke-pixel ;  \ after turtle has moved on (if penup)
: erase-xy      ( x y -- x y )  2dup xy2lin BL white white fgbg2col poke-pixel ;  \ after turtle has moved on ( if penup and backspace )

\ === Show/Hide Turtle
: turtle-hide       ( x y -- x y )  set-hidden  restore-pixel ;   \ overwrite turtle at current location with cursor color, hiding him, but don't change state
: turtle-show       ( x y -- x y )  2dup get-heading show-heading set-visible ; 
: turtle-hide-temp  ( x y -- x y )  restore-pixel ;   \ hide, but don't change state
: visibility-reset  ( x y -- x y )  is-visible? if turtle-show else turtle-hide then ;
: visibility-toggle ( x y -- x y )  is-visible? if turtle-hide else turtle-show then ; 
: cloak-check ( -- ) is-cloaking-on? if else set-visible then ;

\ === Turtle Penmode changes
: do-penup     ( -- ) is-penup? if else  set-penup  save-prev-pen-color  switchto-penup-color  then ; \ pen-up mode has custom appearance;
: do-pendown   ( -- ) clear-penup  restore-prev-pen-color  cloak-check ;
: penup-reset    ( -- ) is-penup? if do-penup else do-pendown then ; 
: penup-toggle   ( -- ) is-penup? if do-pendown else do-penup then ; 

\ Screenshot pixel by pixel.  Each pixel is a byte in memory, but occupies two bytes on screen.
: pic-scr2mem  ( pic-ptr -- ) num-pix-scr 0 do i 2 * ( pic linloc ) peek-pixel rot pix!+ loop drop ;   \ save picture from screen to memory
: pic-mem2scr  ( pic-ptr -- ) num-pix-scr 0 do i 2 * swap pix@+ >r poke-pixel r> loop drop ;           \ load picture from memory to screen
: save-mem2file ( -- )  turtle-hide-temp  my-picture  num-pix-scr  fscr-save  file-write  visibility-reset  ;
: quit-mem2file ( -- )  turtle-hide-temp  picture     num-pix-scr  fscr-quit  file-write  visibility-reset  ;
\ load if files exist
: read-saved ( -- ) ;
: read-quitted ( -- ) ;

: save-screen ( -- ) picture pic-scr2mem ;
: screen-restore ( -- ) picture pic-mem2scr ;                      \ restore screen to remove prompt


\ === Text in Graphics mode
\ ??? CODE-REVIEW - combine these three
: textbox   ( c-addr nchar x y -- ) xy2lin -rot red yellow write-string ;
: promptbox ( c-addr nchar -- )     5 10 xy2lin -rot red lime write-string ;   \ display prompt on screen
: screen-prompt ( c-addr nchar -- ) save-screen promptbox ;         \ save screen before prompting

: prompt-clear     ( -- c-addr nchar ) " Clear screen? (press y or ESC to clear, or any key to continue drawing.)" ;
: prompt-quit      ( -- c-addr nchar ) " Quit? (press y or q to quit or any key to continue drawing)"  ;
: prompt-overwrite ( -- c-addr nchar ) " Overwrite existing skill?  (y/n)" ;
: prompt-full      ( -- c-addr nchar ) " Memory full! (500 instructions max). (Press any key...)" ;
: prompt-full2     ( -- c-addr nchar ) " ... try teaching the rest in the next skill.  (Press any key)" ;
: clear? ( -- t/f )                                     \ y,Y,ESC are confirmation
    prompt-clear screen-prompt
    'y' 'Y' esc-key   key   >r r@ = swap r@ = or swap r> = or               \ check against acceptable values
    if true else false then screen-restore ;
: confirm-quit? ( -- t/f )  prompt-quit  screen-prompt
    'y' 'Y' 'q' 'Q'   key   >r r@ = swap r@ = or swap r@ = or swap r> = or  \ check key against confirming values
    if true else false then 
    screen-restore ;    \ screen snapshot is used in turtle-restore
: overwrite? ( -- t/f ) 
    prompt-overwrite screen-prompt
    'y' 'Y' key dup rot = swap rot = or 
    if true else false then screen-restore ;
: full! ( -- )  prompt-full screen-prompt key drop screen-restore  
                prompt-full2  screen-prompt key drop screen-restore ;


\ === TURTLE MOVEMENT CONTROL
\ movement is either forward or backwards along the axis of the current heading, with screen wrap-around
\ moving forward paints, moving backward erases
\ coordinates (x,y) are semi-cartesian: over and down.  (0,0) is the top left of the user space, i.e. excluding the screen border and menus.
\ movement = 1) update coordinates to a new pixel, 2) paint or restore previous pixel, 3) new pixel is cached, 4) show turtle in new pixel.

\ screen wrap-around
: wrap-y ( y -- y* ) user-height mod ;
: wrap-x ( x -- x* ) user-width mod ;

\ movement along direction axes
: move-N-fwd   ( x y -- x' y' ) 1 - wrap-y ;
: move-N-back  ( x y -- x' y' ) 1 + wrap-y ;
: move-E-fwd   ( x y -- x' y' ) swap 1 + wrap-x swap ;
: move-E-back  ( x y -- x' y' ) swap 1 - wrap-x swap ;
: move-S-fwd   ( x y -- x' y' ) move-N-back ;
: move-S-back  ( x y -- x' y' ) move-N-fwd ;
: move-W-fwd   ( x y -- x' y' ) move-E-back ;
: move-W-back  ( x y -- x' y' ) move-E-fwd ;
: move-NE-fwd  ( x y -- x' y' ) move-N-fwd move-E-fwd ;
: move-NE-back ( x y -- x' y' ) move-N-back move-E-back ;
: move-SW-fwd  ( x y -- x' y' ) move-NE-back ;
: move-SW-back ( x y -- x' y' ) move-NE-fwd ;
: move-NW-fwd  ( x y -- x' y' ) move-N-fwd move-W-fwd ;
: move-NW-back ( x y -- x' y' ) move-N-back move-W-back ;
: move-SE-fwd  ( x y -- x' y' ) move-NW-back ;
: move-SE-back ( x y -- x' y' ) move-NW-fwd ;

\ create motion array of movements in h-index (rotation cycle) order
: create-fwd-motions
create ( | <name> -- )     
  ['] move-N-fwd   ,   ['] move-NE-fwd  ,    ['] move-E-fwd   ,   ['] move-SE-fwd  ,  
  ['] move-S-fwd   ,   ['] move-SW-fwd  ,    ['] move-W-fwd   ,   ['] move-NW-fwd  , 
does> ( x y -- x' y' ) get-heading cells + @ execute ;
: create-back-motions
create ( | <name> -- )     
  ['] move-N-back   ,   ['] move-NE-back  ,    ['] move-E-back   ,   ['] move-SE-back  ,  
  ['] move-S-back   ,   ['] move-SW-back  ,    ['] move-W-back   ,   ['] move-NW-back  , 
does> ( x y -- x' y' ) get-heading cells + @ execute ;
create-fwd-motions    MOVE-FWD-HEADINGS   ( x y -- x' y' )
create-back-motions   MOVE-BACK-HEADINGS  ( x y -- x' y' )

\ if cloaking is off, then every move or heading change makes turtle visible
: move-fwd ( x y -- x' y' )   is-penup? if restore-pixel else paint-xy then  move-fwd-headings  cloak-check ;
: move-back ( x y -- x' y' )  is-penup? if restore-pixel else erase-xy then  move-back-headings cloak-check ;
: change-heading ( 1/-1 -- )  get-heading + 8 mod set-heading cloak-check ;    \ nudges 45 degrees cw (positive) or ccw (negative)

\ display text in direction of turtle heading
: display-char  ( x y asc -- x y )  >r 2dup xy2lin r> get-turtle-color get-pen-color fgbg2col poke-pixel move-fwd-headings cache-pixel ;


\ === RECORDING STATE VARIABLES
create rec-state 1 bytes allot \ States describing status of Recording (on or off or full etc.)
: set-rec-state  create ( n -- ) c,  does> ( -- ) c@ rec-state c! ;
0 set-rec-state rec-ready
1 set-rec-state rec-recording
2 set-rec-state rec-full
3 set-rec-state rec-reminder
4 set-rec-state rec-paused
: get-rec-state ( -- r ) rec-state c@ ;

\ Record strings
5       constant   num-recmsgs                                \ number of different recording messages
12      constant   len-recmsg                                 \ msg length
11      constant diary-numpos                                 \ char position in string for diarynum

\ We show what state the teaching/learning is and dynamically which diary will record the learnings  (see get-rec-str)
: create-recording-names                                    \ holds recording strings and produces them on demand
create ( -- ) here num-recmsgs len-recmsg * bytes allot
  " TeachSkill  "     attach+   " LEARNING    "     attach+
  " MemoryFull  "     attach+   " LEARNING    "     attach+
  " Paused      "     attach+
  drop
does> ( idx base-adr -- c-addr nchar )  len-recmsg  rot get-fw-string ; \ return idx-th fixed width string
create-recording-names REC-MSG-NAMES


\ === KEY LOGGER / DIARY FUNCTION
\ For nested macros to work, we need a diary stack.

9 constant num-diaries                          \ number of distinct recordable macros
500 bytes constant diary-length                 \ max keypress codes in each macro  (frugality breeds resourcefulness, efficiency)
num-diaries diary-length * constant diary-array-sz

create d-index 1 bytes allot                  \ VARIABLE - holds active diary index (0 to 8)
: set-diary-index ( dx -- ) d-index c! ;
: get-diary-index ( -- dx ) d-index c@ ;
: next-diary ( -- ) get-diary-index 1 + num-diaries mod  ( d-index ) set-diary-index ;

: diary-num2dx ( asc --  dx ) a2d 1 - ;       \ from user diary number to coded diary index
: diary-dx2num ( dx  -- asc ) 1 + d2a ; 

create   diary  diary-array-sz bytes allot      \ allocate diary memory
diary diary-array-sz 0 fill                     \ zero out diary memory

create diary-end num-diaries cells allot                   \ ARRAY - holds diary ends to protect against memory overflow when writing (triggers Full! msg)
diary-end num-diaries cells 0 fill                         \ zero out mem
: set-diary-end ( dx dptr -- ) swap cells diary-end + ! ;  \ store position 
: get-diary-end ( dx -- dptr ) cells diary-end + @ ;
: diary-full?  ( dx dptr -- t/f ) 2dup swap get-diary-end > ;

create diary-pos num-diaries cells allot  \ ARRAY - holds next free place in each diary, which is the stop point for playback
diary-pos num-diaries cells 0 fill      \ zero out mem
: set-diary-pos ( dx dptr -- )  swap cells diary-pos + ! ;                           \ store position
: get-diary-pos ( dx -- dptr )  cells diary-pos + @ ;

: get-diary-start ( dx -- dx dptr )  dup diary diary-length rot * + ;                 \ returns start position
: clear-diaries ( -- )
    0 get-diary-start num-diaries 0 do  ( dx dptr ) 
       2dup set-diary-pos                     \ set start
       diary-length + 2dup 1 - set-diary-end  \ set end
       ( dx dptr ) swap 1 + swap              \ next diary
    loop 2drop ;
: clear-diary ( dx -- ) get-diary-start 2dup set-diary-pos diary-length 0 fill drop ;


\ === PAINT TO MENU
\ !! HOW IT SHOULD WORK (how it works seems more complicated)
\ The menu display system should be simple: draw the menu, refresh dynamic content when it changes.
\ Do this to a virtual copy of the menu so that it persists even with screen changes (e.g. display help system, clear screen, etc.)
\ Text should use the default fgbg unless in a special mode, in which case this is provided instead.
\ !! END CODE REVIEW

\ Menu parameters: 
\ y-screen location of menus  (if change these, also change top-, bot-border-row)
22  constant menu2-row     \ bottom menu (y=.)
23  constant menu3-row     \ mid menu
24  constant menu1-row     \ top menu (y=.)
\ x-screen coordinate of dynamic content in menu2
5   constant menu2-h#      \ heading
18  constant menu2-c#      \ color
29  constant menu2-r#      \ recording status
6   constant menu3-t#      \ text mode
12  constant menu3-d#      \ draw mode
26  constant menu3-z#      \ cloaking mode

: menu2-text ( -- )  " Hdng:-- [X/C]olor:------ [T]:------------ [N]extSkill [1-9]Play-/[0]List-Skills " ;     
: menu3-text ( -- )  " Mode:[Text ]Draw [Z]Cloak:Off [Esc]ClearScr [S/U]Save/Undo [;]Show/HideTurtle   " ;          \ [L]oop:--
: menu1-text ( -- )  " [Arrows]MoveFwd,TurnL/R [Bkspc]Erase [']Penup/dwn [H]ome [G]reet [?]HELP [Q]UIT " ;
\                      ********************************************************************************   \ template - max 80 char

\ === MENU WRITING CODE ==  ( linloc c-adr nchar fg bg -- )
: test-setup 0 15 xy2lin-screen menu1-text red lime ;     \ 0 15 are row col (in screen coords) for starting the string write
: test test-setup write-string ;

\ foreground/background color for text
: menu1-fgbg  white navy ;
: menu2-fgbg  white navy ;
: menu3-fgbg  white navy ; 
: menu2-heading-position ( -- loc )  menu2-h# menu2-row  ;
: menu2-color-position   ( -- loc )  menu2-c# menu2-row  ;  
: menu2-rec-pos          ( -- loc)   menu2-r# menu2-row ;
\ ??? CODE REVIEW -- SIMPLIFY
: menu2-dynamic-fgbg ( -- ) navy lime ; \ color scheme to accent dynamic text
: menu2-attn-fgbg    ( -- ) red lime ;  \ color scheme to highlight text
: create-rec-fgbg      \ based on rec state
    create 
        navy c, lime c,     \ if 0 
        red c, yellow c,      \ if 1
        navy c, yellow c,     \ if 2
        white c, red c,       \ if 3
        white c, orange c,    \ if 4
    does> ( -- )
       get-rec-state 2 * bytes + c@+ swap c@ ; 
create-rec-fgbg menu2-rec-fgbg

\ creating menus
: new-menu 
    create ( menu-text_c-addr|nchar  menu-row  fg bg  | <name> -- )   \ create dictionary entry
    fgbg2col c,                                                       \ create default menu color code
    physical-cols * ,                                                 \ physical offset to start on screen at (0,menu-row)
    here dup , screen-cols 2dup dup bytes allot BL fill drop swap cmove   \ allocate memory for string, initialized with blanks, then copied with menu-text
does>  ( -- )                                                          \  places address of entry on stack 
    c@+ swap  \ fgbg
    @+ swap   \ start-loc physical   
( col L ca )
    screen-cols 0 do write-char+ loop                                  \ print menu to screen
    3drop ;

menu1-text  menu1-row  menu1-fgbg   new-menu  menu1
menu2-text  menu2-row  menu2-fgbg   new-menu  menu2
menu3-text  menu3-row  menu3-fgbg   new-menu  menu3

: test-pixel  1000 'X' 31 poke-pixel ;
: setup-menutest menu2-text menu2-row menu2-fgbg ;
: test-create ( <menu3> ) setup-menutest new-menu ;

\ dynamically updating menus
: menu-write ( c-addr length fg bg xcol-screen yrow-screen -- )
    xy2lin-screen             \ calc physical offset
    -rot >r >r -rot r> r>     \ prepare stack
    ( linloc c-addr nchar fg bg )
    write-string ;
: test-menu-write  5 color-names red yellow 20 23 menu-write ;

\ dynamically update diary-string in menu
: update-rec-str  ( c-addr nchar -- c-addr nchar )  2dup drop get-diary-index diary-dx2num swap diary-numpos + c! ;   \ updates string with diary number
: get-rec-str ( -- c-addr nchar )  get-rec-state  rec-msg-names  update-rec-str  ;                       \ returns valid recmsg

: menu2-show-heading  ( -- )              get-heading  heading-string  menu2-dynamic-fgbg  menu2-h#  menu2-row  menu-write ;
: get-pencolor-name   ( -- c-adr nchar )  is-penup?    if penup-col-index else get-pen-color then color-names ;
: get-color-notification     ( -- fg bg ) is-penup?    if menu2-attn-fgbg else menu2-dynamic-fgbg then ;
: menu2-show-color           ( -- )       get-pencolor-name get-color-notification menu2-c# menu2-row menu-write ;
: menu2-show-recording-state ( -- )       get-rec-str  menu2-rec-fgbg  menu2-r# menu2-row menu-write ;
: menu3-show-mode            ( -- )       is-text-mode? 
    if " Text" menu2-attn-fgbg menu3-t# menu3-row menu-write
       " Draw" menu2-fgbg menu3-d# menu3-row menu-write
    else " Draw"  menu2-attn-fgbg     menu3-d# menu3-row menu-write
         " Text"  menu2-fgbg menu3-t# menu3-row menu-write 
    then ;
: menu3-cloak-mode           ( -- )  is-cloaking-on? 
    if    " On "  menu2-attn-fgbg  menu3-z# menu3-row menu-write 
    else  " Off"  menu2-fgbg       menu3-z# menu3-row menu-write 
    then ;
    
: paint-static-menus ( -- )   menu1 menu3 menu2 ;  \ paints static menus
: refresh-menus      ( -- )   menu2-show-heading  menu2-show-color  menu2-show-recording-state  menu3-show-mode  menu3-cloak-mode ; \ repaints dynamic portions
: paint-screen       ( -- )   get-screen-color fill-screen   borders-draw   paint-static-menus ;


\ === VIEWING THE DIARY
8 constant octet
3 constant triple
octet triple * constant diary-row-size
: show-octet ( -- ) octet 0 do space c@+ emit loop ;
: show-row ( dptr -- dptr' ) cr triple 0 do show-octet space space loop ; 
: cue-to-start ( dx -- dptr ) diary diary-length rot * +  ;  \ cue to diary start position
: fetch-end    ( dx -- dptr ) cells diary-pos + @ ; 
: show-max-page ( nrows -- ) 0 do show-row i 1 + 20 mod 0 = if pause-key then loop ;  \ show one screenful at a time
: dump-diary ( dx -- )  \ show entire diary, filled or not
    cue-to-start diary-length diary-row-size / show-max-page drop ;
: see-one-diary ( dx -- ) 
    dup cue-to-start swap fetch-end over - diary-row-size / 1 + show-max-page drop ;
: see-diaries ( -- )   \ pause after each diary view
    9 0 do cr ." Skill #" i dup diary-dx2num emit ." : " see-one-diary cr pause-esc if leave then loop cr ;
    
\ ??? CODE REVIEW --- THESE are not linked to the recording state variables --- design flaw - combine.
create diary-recording 1 cells allot   \ BOOLEAN VARIABLE - true if recording
: set-diary-recording ( -- ) true diary-recording c! ;
: clear-diary-recording ( -- ) false diary-recording c! ;
: is-recording? ( -- t/f ) diary-recording c@ ;


\ === DIARY STACK - permits nested macros
create diary-stack 16 bytes allot               \ STACK - saves context for nested macros
create tods         1 cells allot               \ VARIABLE - holds top of diary stack
: set-tods ( dstack-ptr --  ) tods ! ; 
: get-tods ( -- dstack-ptr  ) tods @ ; 

: init-diary-stack diary-stack dup  16  $FF fill set-tods ;   \ set memory to $FF and set top of stack
init-diary-stack

: push-diary ( dx -- ) get-tods c!+ set-tods ;
: pop-diary ( -- dx ) tods @ 1 - dup c@ swap set-tods ;

: dump-diary-stack ( -- ) diary-stack 16 0 do c@+ . loop drop ;
: see-diary-stack ( -- ) tods @ 1 + diary-stack do i c@ . loop ;
: test-diarystack ( 3 2 1 0 -- )   \ run it on command line
    dump-diary-stack see-diary-stack 
    3 2 1 0 push-diary push-diary push-diary push-diary 
    dump-diary see-diaries 
    pop-diary . pop-diary . pop-diary . ;


\ === DIARY RECORDINGS
: show-recording ( -- )  set-diary-recording  rec-recording  menu2-show-recording-state ;
\ ??? CODE-REVIEW - this changes recording state, i.e. it flashes and then SETS recording state.  Should flash and return to previous state.
: flash-recording ( -- )  rec-reminder  menu2-show-recording-state 300 ms show-recording ;

: ok-to-write?     ( dx -- dx t/f ) get-diary-start over get-diary-pos = if true else overwrite? then ;           \ checks if diary has content
: start-recording  ( -- )           get-diary-index  ok-to-write?  if clear-diary  show-recording else drop then  ;  \ record macro from beginning if ok 
: resume-recording ( -- )           show-recording ;                                                              \ record macro from previous position
: stop-recording   ( -- )           clear-diary-recording    rec-ready   menu2-show-recording-state ;
: pause-recording  ( -- )           clear-diary-recording    rec-paused  menu2-show-recording-state ; 

: warn-diary-full ( -- )  rec-full  menu2-show-recording-state  full!  stop-recording ;
: diary-write-entry ( dx asc -- )  over get-diary-pos c!+ ( dx dptr' )  diary-full? if 2drop warn-diary-full else set-diary-pos then ;

                                
\ ============ Control keys
\ Keyboard commands are associated with their ASCII codes.
\ Design is child friendly, with upper and lower case letters mapped to the same command so the program works even if caps lock gets pressed.
\    / is mapped to help (?) since ? requires use of SHIFT which is not easy for a child
\    (numbers are not affected by Caps Lock)
\ Note: added printable characters for all commands so can record in diary.  Arrows still work, but as synonymns.  Primary (recorded) commands are printable.
\ The primary key will get recorded into the diary, not all variations
$C8  constant  fwd-arrow     'F'  constant  fwd-key      'f'  constant  fwd-key2      fwd-arrow  constant  fwd-key3       \ $C8=200d
$CD  constant  right-arrow   'R'  constant  right-key    'r'  constant  right-key2    right-arrow  constant  right-key3    right-key  constant  cw-key \ 205d
$CB  constant  left-arrow    'L'  constant  left-key     'l'  constant  left-key2     left-arrow   constant  left-key3     left-key   constant  ccw-key \ 203d
$8   constant  backspace     'B'  constant  erase-key    'b'  constant  erase-key2    backspace    constant    erase-key3
$27  constant    penupdown-key      \ tick
'c'  constant    color-fwd-key      'C'  constant    color-fwd-key2     \ change color
'x'  constant    color-back-key     'X'  constant    color-back-key2    \ change color
'n'  constant    diary-key          'N'  constant    diary-key2         \ RENAMED NEXT
't'  constant    record-key         'T'  constant    record-key2        \ RENAMED TEACH
'1'  constant    play-1-key    '2'  constant    play-2-key    '3'  constant    play-3-key    '4'  constant    play-4-key
'5'  constant    play-5-key    '6'  constant    play-6-key    '7'  constant    play-7-key    '8'  constant    play-8-key
'9'  constant    play-9-key
'0'  constant    show-diary-key
's'  constant    save-key          'S'  constant    save-key2
'u'  constant    undo-key          'U'  constant    undo-key2
';'  constant    show-hide-key      \ turtle hide/view 
'h'  constant    home-key          'H'  constant    home-key2
esc-key  constant    clear-screen-key
'?'  constant    help-key      '/'      constant    help-key2           \ Usability: in case a child does not know how to use shift key
'q'  constant    quit-key      'Q'      constant    quit-key2
'g'  constant    greeting-key  'G'      constant    greeting-key2
'['  constant    left-bracket
']'  constant    right-bracket
'z'  constant    cloak-key     'Z'      constant    cloak-key2


\ === KEY PROCESSING ENGINE ===
\ Turtle responds to user keypresses using a key mapping lookup table which associates every ascii key with the executable token (xt) giving its intended action (if it has one).
\ Mapping table is defined below (after all actions are defined) - search =keymap=
create   keymap   256 cells allot                         \ creates storage for executable tokens (xt) to be logged at their ascii code offset location

: create-mapping  ( keymap <name> -- )   \ compiles keymap base address into <named> instruction.  Generic design can be used for multi-language keymaps
create ,                                 
does> ( x y asc -- x y )                 \ key processing engine: key commands must consume ascii keycode and preserve x,y
    @ ( keymap)  over cells + @ ( x y asc xt ) execute ( x y )  refresh-menus  visibility-reset  ;
keymap create-mapping    TURTLE-DO-KEY


\ === TURTLE PROCESSING COMMANDS (by TURTLE-DO-KEY) ===
: is-current-diary? ( macro-asc -- macro-asc t/f ) dup diary-num2dx get-diary-index = ; 
: record-asc2diary ( asc -- asc ) get-diary-index over ( asc dx asc ) diary-write-entry ; 
: log-key-if-recording ( asc -- asc ) is-recording? if record-asc2diary then ;  \ required if command is to be logged

\ These commands are logged if recording is active ( normalized character key is logged, not necessarily the key that was pressed )
: penupdown ( x y asc -- x y ) log-key-if-recording  drop penup-toggle ;

: turtle-move-fwd  ( x y asc -- x' y' ) drop fwd-key   log-key-if-recording drop move-fwd  cache-pixel ;
: turtle-move-back ( x y asc -- x' y' ) drop erase-key log-key-if-recording drop move-back cache-pixel ;
: heading-change-left  ( asc -- )       drop left-key  log-key-if-recording drop -1 change-heading  ;  \ nudge 1 click ccw
: heading-change-right ( asc -- )       drop right-key log-key-if-recording drop  1 change-heading  ;  \ nudge 1 click cw

: color-change-fwd  ( x y asc -- x y )  \ nudge 1 color fwd.  \ changing color forces pendown mode  
    drop color-fwd-key  log-key-if-recording  drop 
    is-penup? if do-pendown 1 color-change-ex save-prev-pen-color else 1 color-change-ex then ;
: color-change-back ( x y asc -- x y ) 
    drop color-back-key  log-key-if-recording  drop 
    is-penup? if do-pendown -1 color-change-ex save-prev-pen-color else -1 color-change-ex then ;

: greeting      ( x y asc -- x y )  log-key-if-recording drop save-screen 2dup 1 + xy2lin    \ show text on next row down
    " Hello!" red yellow write-string 300 ms screen-restore ;  \ briefly flash hello

\ These commands are SOMETIMES logged if recording is active.
: macro-playback ( x y asc -- x' y' )  \ show macro playing on screen
    diary-num2dx  dup set-diary-index  get-diary-start
    swap get-diary-pos swap
    2dup <= if 2drop  else  do i c@ turtle-do-key loop  then ;

: play-macro ( x y macro-asc -- x y )
    is-recording?  
if  b
    is-current-diary?    \ log key if not current diary: nested macros allowed (mixture of primary and derived commands) but not recursion (calling oneself)
    if flash-recording  drop  \ recursion (calling onself) not allowed - warning and no playback 
    else 
       record-asc2diary                \ record macro-name not individual keypresses
       pause-recording                 \ playback individual keypresses but without recording       
       get-diary-index  push-diary       \ diary stack
       macro-playback                  \ want to see the macro playing on the screen as constructing the new macro
       pop-diary set-diary-index       \ store diary context from the stack
       resume-recording                \ allow recording to continue from same position
    then
else 
    macro-playback                  \ replay macro
then ;

\ These commands are not logged (recording is flashed as reminder)
: save-point ( x y asc -- x y )                                   \ save screenshot without turtle as a restore point
    drop is-recording? if flash-recording then  save-xy 
    turtle-hide-temp  my-picture  pic-scr2mem  save-mem2file  ;   \ hide turtle before capture screenshot
: restore-point ( x y asc -- x y ) drop is-recording? if flash-recording else my-picture pic-mem2scr then restore-xy ;  \ restore to last saved image and replacing turtle coords
: undo  restore-point ;   \ synonymm

: cloak        ( -- )  enable-cloaking   save-prev-penup-state  do-penup   ;
: uncloak      ( -- )  disable-cloaking  restore-prev-penup-state  penup-reset ; 
: cloak-toggle ( -- )  is-cloaking-on?  if  uncloak  turtle-show  else  cloak  turtle-hide then ;
: cloaking-onoff ( x y asc -- x y ) drop is-recording? if flash-recording else cloak-toggle then ;  \ toggle cloaking functionality

: turtle-showhide ( x y asc -- x y ) drop is-recording? if flash-recording else visibility-toggle then ;

\ These commands are not logged (no visible indication) and recording is stopped when they are pressed.
: startup-turtle ( -- x y )  init-heading  init-turtle-position  cache-pixel ;
: turtle-home ( x y -- x y )  \ preserves turtle states, don't leave mark in last position & don't clobber home position
    turtle-hide-temp  save-prev-penup-state  do-penup  2drop  startup-turtle  restore-prev-penup-state  penup-reset ;
: go-home ( x y asc -- x y ) drop is-recording? if stop-recording then turtle-home ;    \ reset turtle to home position

: recording-onoff ( x y asc -- x y )  drop is-recording? if stop-recording else start-recording then ;
: diary-next      ( x y asc -- x y )  drop is-recording? if stop-recording then next-diary ;


\ These commands are not allowed while recording
: is-quit? ( asc -- asc t/f ) 
        quit-key over <> over quit-key2 <> and  ( asc t/f )     \ if not q and not Q
        if false                                                \ then don't quit
        else is-recording? if flash-recording false             \ can't quit while recording
            else confirm-quit? then                             \ check that actually want to quit - confirm-quit? snapshots screen (used by turtle-resume)
        then ;
: clear-screen ( x y asc -- x y )  drop
        is-recording? if stop-recording then
        clear?   if  turtle-home   paint-screen  cache-pixel  then  ;

\ These commands are not logged but do not interrupt recording which resumes afterwards
: show-help   ( -- )             save-screen  cr black-screen  welcome  pause-key  cr screen-restore ; 
: help-screen ( x y asc -- x y ) drop show-help ;
: show-diary  ( x y asc -- x y ) drop save-screen see-diaries screen-restore ;

\  UNDECIDED logging behaviour
: text-mode-on  ( x y asc -- x y )  drop set-text-mode do-pendown ; \ starts text mode brings pen down.


\ === KEY PROCESSING ENGINE (part 2) ===
\ Actions (responses) have their executable tokens saved into KEYMAP at their ascii keycode offset.  Non mapped keys are ignored by default
: add-key       ( xt keycode -- ) cells keymap + ! ;     \   stores response (xt) into keymap table at keycode offset location
: ignore-keypress  ( asc -- )        drop ;    
: init-keymap   ( -- )            ['] ignore-keypress 255 0 do dup i add-key loop drop ;    \ initialize every key in keymap to ignore input; 
: map-key-commands ( -- )         init-keymap       \ overwrite mapped keys with action/response
\         xt               keycode
  ['] penupdown                 penupdown-key      add-key
  ['] turtle-showhide           show-hide-key      add-key
  ['] text-mode-on              left-bracket       add-key   \ text mode is turned off w/ right-bracket or various exit keys, checked within text mode loop
  ['] show-diary                show-diary-key     add-key
  ['] clear-screen              clear-screen-key   add-key
  ['] go-home               dup home-key           add-key       home-key2       add-key
  ['] diary-next            dup diary-key          add-key       diary-key2      add-key
  ['] recording-onoff       dup record-key         add-key       record-key2     add-key
  ['] save-point            dup save-key           add-key       save-key2       add-key
  ['] restore-point         dup undo-key           add-key       undo-key2       add-key
  ['] help-screen           dup help-key           add-key       help-key2       add-key
  ['] greeting              dup greeting-key       add-key       greeting-key2   add-key
  ['] cloaking-onoff        dup cloak-key          add-key       cloak-key2      add-key
  ['] color-change-fwd      dup color-fwd-key      add-key       color-fwd-key2  add-key   
  ['] color-change-back     dup color-back-key     add-key       color-back-key2 add-key
  ['] turtle-move-fwd       dup fwd-key            add-key   dup fwd-key2        add-key       fwd-key3    add-key
  ['] turtle-move-back      dup erase-key          add-key   dup erase-key2      add-key       erase-key3  add-key
  ['] heading-change-left   dup left-key           add-key   dup left-key2       add-key       left-key3   add-key
  ['] heading-change-right  dup right-key          add-key   dup right-key2      add-key       right-key3  add-key
  ['] play-macro            dup play-1-key         add-key   dup play-2-key      add-key   dup play-3-key  add-key   dup play-4-key add-key
    dup play-5-key add-key  dup play-6-key         add-key   dup play-7-key      add-key   dup play-8-key  add-key       play-9-key add-key
;


\ === MAIN PROGRAM ===
: turtle-init ( -- x y )   \ initialization code
    map-key-commands  clear-diaries  disable-cloaking
    text-video-mode   red set-pen-color   save-prev-pen-color  silver set-border-color   white set-screen-color   paint-screen   
    stop-recording    set-draw-mode  set-visible   do-pendown   0 set-diary-index   
    startup-turtle    refresh-menus   save-xy   save-screen  my-picture pic-scr2mem   turtle-show ;
: stop-char?    ( asc -- asc t/f )  dup is_char? not over ']' = or ;   \ non-chars or ] signal stop
: process-text-loop ( x y -- x y ) 
       begin ( x y ) key stop-char? ( x y asc t/f )                \ fetch keys until stop is signalled 
        dup if over erase-key3 = 
             if not else set-draw-mode refresh-menus then          \ backspace stays in text mode, all other non-char keys signal exit
             >r turtle-do-key r>                                   \ stop-chars are processed as turtle instructions
        else >r display-char r>                                    \ chars are displayed
        then
       until ;
: process-key-loop ( x y -- x y )                                  \ fetch keys until quit, entering separate text processing loop if text mode is selected
    begin  key  ( asc )  is-quit?
      if drop true  else  turtle-do-key 
          is-text-mode? if process-text-loop then 
          false 
      then 
    until ;
: show-quit-message ( -- )  cr ." To restart, type TURTLE"  cr ." To return to Windows, type EXIT." cr ;
: turtle-quit    ( -- ) is-recording? if stop-recording  then  quit-mem2file  show-quit-message ( bye ) ;  \ exit Forth to DOS (except if bye is commented out)
: turtle-run     ( x y -- x y )  process-key-loop        turtle-quit ;
: turtle-resume  ( x y -- x y )  screen-restore          turtle-run  ; 
: turtle-go      ( -- x y )      show-help  turtle-init  turtle-run  ;    \ start application
\ === END PROGRAM ===
